import os
import re
import tempfile
import datetime as dt

import pandas as pd
from pandas.tseries.offsets import MonthEnd
import patoolib
import requests
from bs4 import BeautifulSoup

from econuy.config import ROOT_DIR
from econuy.processing import updates, columns
from econuy.resources.utils import fiscal_url, fiscal_sheets

DATA_PATH = os.path.join(ROOT_DIR, "data")
update_threshold = 25


def get(update: bool = False, revise_rows: int = 0,
        save: bool = False, force_update: bool = False):
    """Get fiscal data.

    Parameters
    ----------
    update : bool (default is False)
        If true, try to update existing data on disk.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save : bool (default is False)
        If true, save output dataframe in CSV format.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for fiscal data, 25 days)

    Returns
    -------
    output : dictionary of Pandas dataframes
        Each dataframe corresponds to a sheet in the fiscal accounts Excel
        provided by the MEF. Not all sheets are considered.

    """
    if update is True:
        update_path = os.path.join(DATA_PATH, "fiscal_nfps.csv")
        modified = dt.datetime.fromtimestamp(os.path.getmtime(update_path))
        delta = (dt.datetime.now() - modified).days

        if delta < update_threshold and force_update is False:
            print(f"Fiscal data was modified within {update_threshold} day(s)."
                  f" Skipping download...")
            output = {}
            for metadata in fiscal_sheets.values():
                update_path = os.path.join(DATA_PATH,
                                           metadata['Name'] + ".csv")
                delta, previous_data = updates.check_modified(update_path)
                output.update({metadata["Name"]: previous_data})
            return output

    response = requests.get(fiscal_url)
    soup = BeautifulSoup(response.content, "html.parser")
    links = soup.find_all(href=re.compile("\\.rar$"))
    rar = links[0]["href"]

    temp_rar = tempfile.NamedTemporaryFile(suffix=".rar").name
    with open(temp_rar, "wb") as f:
        f.write(requests.get(rar).content)

    with tempfile.TemporaryDirectory() as temp_dir:
        patoolib.extract_archive(temp_rar, outdir=temp_dir)
        path = os.path.join(temp_dir, os.listdir(temp_dir)[0])

        output = {}
        with pd.ExcelFile(path) as xls:
            for sheet, metadata in fiscal_sheets.items():
                if update is True:
                    update_path = os.path.join(DATA_PATH,
                                               metadata['Name'] + ".csv")
                    delta, previous_data = updates.check_modified(update_path)

                    if delta < update_threshold and force_update is False:
                        print(f"{metadata['Name']}.csv was modified within "
                              f"{update_threshold} day(s). "
                              f"Skipping download...")
                        output.update({metadata["Name"]: previous_data})
                        continue

                data = (pd.read_excel(xls, sheet_name=sheet).
                        dropna(axis=0, thresh=4).dropna(axis=1, thresh=4).
                        transpose().set_index(2, drop=True))
                data.columns = data.iloc[0]
                data = data[data.index.notnull()].rename_axis(None)
                data.index = data.index + MonthEnd(1)
                data.columns = metadata["Colnames"]

                if update is True:
                    data = updates.revise(new_data=data,
                                          prev_data=previous_data,
                                          revise_rows=revise_rows)
                data = data.apply(pd.to_numeric, errors="coerce")
                columns.set_metadata(
                    data, area="Cuentas fiscales y deuda", currency="UYU",
                    inf_adj="No", index="No", seas_adj="NSA", ts_type="Flujo",
                    cumperiods=1
                )

                if save is True:
                    save_path = os.path.join(DATA_PATH,
                                             metadata["Name"] + ".csv")
                    data.to_csv(save_path, sep=" ")

                output.update({metadata["Name"]: data})

    return output



import os
import datetime as dt
from pathlib import Path
from typing import Union

import pandas as pd
import numpy as np
import requests
from pandas.tseries.offsets import MonthEnd

from econuy.retrieval import cpi, nxr
from econuy.config import ROOT_DIR
from econuy.processing import index, updates, columns
from econuy.resources.utils import reer_url, ar_cpi_url, ar_cpi_payload

DATA_PATH = os.path.join(ROOT_DIR, "data")
update_threshold = 25


def official(update: Union[str, Path, None] = None, revise_rows: int = 0,
             save: Union[str, Path, None] = None, force_update: bool = False):
    """Get official real exchange rates from the BCU website.

    Parameters
    ----------
    update : str, Path or None (default is None)
        Path or path-like string pointing to a CSV file for updating.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save : str, Path or None (default is None)
        Path or path-like string where to save the output dataframe in CSV
        format.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for CPI, 25 days)

    Returns
    -------
    proc : Pandas dataframe

    """
    if update is not None:
        update_path = os.path.join(DATA_PATH, update)
        delta, previous_data = updates.check_modified(update_path)

        if delta < update_threshold and force_update is False:
            print(f"{update} was modified within {update_threshold} day(s). "
                  f"Skipping download...")
            return previous_data

    raw = pd.read_excel(reer_url, skiprows=8, usecols="B:H", index_col=0)
    proc = raw.dropna(how="any")
    proc.columns = ["Global", "Regional", "Extrarregional",
                    "Argentina", "Brasil", "EEUU"]
    proc.index = pd.to_datetime(proc.index) + MonthEnd(1)

    if update is not None:
        proc = updates.revise(new_data=proc, prev_data=previous_data,
                              revise_rows=revise_rows)

    columns.set_metadata(proc, area="Precios y salarios", currency="-",
                         inf_adj="No", index="2017", seas_adj="NSA",
                         ts_type="-", cumperiods=1)

    if save is not None:
        save_path = os.path.join(DATA_PATH, save)
        proc.to_csv(save_path, sep=" ")

    return proc


def custom(update: Union[str, Path, None] = None, revise_rows: int = 0,
           save: Union[str, Path, None] = None, force_update: bool = False):
    """Calculate custom real exchange rates from various sources.

    Parameters
    ----------
    update : str, Path or None (default is None)
        Path or path-like string pointing to a CSV file for updating.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save : str, Path or None (default is None)
        Path or path-like string where to save the output dataframe in CSV
        format.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for CPI, 25 days)

    Returns
    -------
    output : Pandas dataframe

    """
    if update is not None:
        update_path = os.path.join(DATA_PATH, update)
        delta, previous_data = updates.check_modified(update_path)

        if delta < update_threshold and force_update is False:
            print(f"{update} was modified within {update_threshold} day(s). "
                  f"Skipping download...")
            return previous_data

    url_ = "http://dataservices.imf.org/REST/SDMX_JSON.svc/CompactData/IFS/M."
    url_extra = ".?startPeriod=1970&endPeriod="
    raw = []
    for country in ["US", "BR", "AR"]:
        for indicator in ["PCPI_IX", "ENDA_XDC_USD_RATE"]:
            base_url = (f"{url_}{country}.{indicator}{url_extra}"
                        f"{dt.datetime.now().year}")
            r_json = requests.get(base_url).json()
            data = r_json["CompactData"]["DataSet"]["Series"]["Obs"]
            try:
                data = pd.DataFrame(data)
                data.set_index("@TIME_PERIOD", drop=True, inplace=True)
            except ValueError:
                data = pd.DataFrame(np.nan,
                                    index=pd.date_range(start="1970-01-01",
                                                        end=dt.datetime.now(),
                                                        freq="M"),
                                    columns=[f"{country}.{indicator}"])
            if "@OBS_STATUS" in data.columns:
                data.drop("@OBS_STATUS", inplace=True, axis=1)
            data.index = (pd.to_datetime(data.index, format="%Y-%m")
                          + MonthEnd(1))
            data.columns = [f"{country}.{indicator}"]
            raw.append(data)
    raw = pd.concat(raw, axis=1, sort=True).apply(pd.to_numeric)

    ar_black_xr, ar_cpi = missing_ar()
    proc = raw.copy()
    proc["AR.PCPI_IX"] = ar_cpi
    ar_black_xr = pd.concat([ar_black_xr, proc["AR.ENDA_XDC_USD_RATE"]], axis=1)
    ar_black_xr[0] = np.where(pd.isna(ar_black_xr[0]),
                              ar_black_xr["AR.ENDA_XDC_USD_RATE"],
                              ar_black_xr[0])
    proc["AR.ENDA_XDC_USD_RATE_black"] = ar_black_xr.iloc[:, 0]
    proc["AR_E_A"] = proc.iloc[:, [5, 6]].mean(axis=1)

    uy_cpi = cpi.get(update="cpi.csv", revise_rows=6,
                     save="cpi.csv", force_update=False)
    uy_e = nxr.get(update="nxr.csv", revise_rows=6,
                   save="nxr.csv", force_update=False).iloc[:, [3]]
    proc = pd.concat([proc, uy_cpi, uy_e], axis=1)
    proc = proc.interpolate(method="linear", limit_area="inside")
    proc = proc.dropna(how="any")
    proc.columns = ["US_P", "US_E", "BR_P", "BR_E", "AR_P", "AR_E",
                    "AR_E_B", "AR_E_A", "UY_P", "UY_E"]

    output = pd.DataFrame()
    output["UY_E_P"] = proc["UY_E"] / proc["UY_P"]
    output["TCR_UY_AR"] = output["UY_E_P"] / proc["AR_E_A"] * proc["AR_P"]
    output["TCR_UY_BR"] = output["UY_E_P"] / proc["BR_E"] * proc["BR_P"]
    output["TCR_UY_US"] = output["UY_E_P"] * proc["US_P"]
    output["TCR_AR_US"] = proc["BR_E"] * proc["US_P"] / proc["BR_P"]
    output["TCR_BR_US"] = proc["AR_E"] * proc["US_P"] / proc["AR_P"]
    output.drop("UY_E_P", axis=1, inplace=True)

    columns.set_metadata(output, area="Precios y salarios", currency="-",
                         inf_adj="-", index="-", seas_adj="NSA",
                         ts_type="Flujo", cumperiods=1)
    output = index.base_index(output, start_date="2010-01-01",
                              end_date="2010-12-31", base=100)

    if update is not None:
        output = updates.revise(new_data=proc, prev_data=previous_data,
                                revise_rows=revise_rows)

    if save is not None:
        save_path = os.path.join(DATA_PATH, save)
        output.to_csv(save_path, sep=" ")

    return output


def missing_ar():
    """Get Argentina's non-official exchange rate and CPI."""
    today_fmt = dt.datetime.now().strftime("%d-%m-%Y")
    black_url = (f"https://mercados.ambito.com/dolar/informal/"
                 f"historico-general/11-01-2002/{today_fmt}")
    black_r = requests.get(black_url).json()
    black_xr = pd.DataFrame(black_r)
    black_xr.set_index(0, drop=True, inplace=True)
    black_xr.drop("Fecha", inplace=True)
    black_xr = black_xr.replace(",", ".", regex=True).apply(pd.to_numeric)
    black_xr.index = pd.to_datetime(black_xr.index, format="%d-%m-%Y")
    black_xr = black_xr.mean(axis=1).to_frame().sort_index()
    black_xr = black_xr.resample("M").mean()

    cpi_r = requests.get(ar_cpi_url, params=ar_cpi_payload)
    cpi_ar = pd.read_html(cpi_r.content)[0]
    cpi_ar.set_index("Fecha", drop=True, inplace=True)
    cpi_ar.index = pd.to_datetime(cpi_ar.index, format="%d/%m/%Y")
    cpi_ar.columns = ["nivel"]
    cpi_ar = cpi_ar.divide(10)

    ps_url = "http://www.inflacionverdadera.com/Argentina_inflation.xls"
    cpi_ps = pd.read_excel(ps_url)
    cpi_ps.set_index("date", drop=True, inplace=True)
    cpi_ps.index = cpi_ps.index + MonthEnd(1)
    cpi_ps = cpi_ps.loc[(cpi_ps.index >= "2006-12-31") &
                        (cpi_ps.index <= "2016-12-01"), "index"]
    cpi_ps = cpi_ps.to_frame().pct_change(periods=1).multiply(100)
    cpi_ps.columns = ["nivel"]

    cpi_all = (cpi_ar.append(cpi_ps).reset_index().
               drop_duplicates(subset="index", keep="last").
               set_index("index", drop=True).sort_index())

    cpi_all = cpi_all.divide(100).add(1).cumprod()

    return black_xr, cpi_all



from datetime import date
from typing import Union

import pandas as pd

from econuy.retrieval import cpi, national_accounts, nxr
from econuy.processing import freqs, columns


def usd(df: pd.DataFrame):
    """Convert dataframe from UYU to USD.

    Convert a dataframe's columns from Uruguayan pesos to US dollars. Call the
    `nxr.get()` function to obtain nominal exchange rates, and take into
    account whether the input dataframe's `Type`, as defined by its multiindex,
    is flow or stock, in order to choose end of period or monthly average NXR.
    Also take into account the input dataframe's frequency and whether columns
    represent rolling averages or sums.

    Parameters
    ----------
    df : Pandas dataframe

    Returns
    -------
    converted_df : Pandas dataframe

    """
    inferred_freq = pd.infer_freq(df.index)
    nxr_data = nxr.get(update="nxr.csv", revise_rows=6,
                       save="nxr.csv", force_update=False)

    if df.columns.get_level_values("Tipo")[0] == "Flujo":
        columns.set_metadata(nxr_data, ts_type="Flujo")
        nxr_freq = freqs.freq_resample(nxr_data, target=inferred_freq,
                                       operation="average").iloc[:, [1]]
        cum_periods = int(df.columns.get_level_values("Acum. períodos")[0])
        nxr_freq = freqs.rolling(nxr_freq, periods=cum_periods,
                                 operation="average")

    else:
        columns.set_metadata(nxr_data, ts_type="Stock")
        nxr_freq = freqs.freq_resample(nxr_data, target=inferred_freq,
                                       operation="average").iloc[:, [3]]

    nxr_to_use = nxr_freq[nxr_freq.index.isin(df.index)].iloc[:, 0]
    converted_df = df.apply(lambda x: x / nxr_to_use)
    columns.set_metadata(converted_df, currency="USD")

    return converted_df


def real(df: pd.DataFrame, start_date: Union[str, date, None] = None,
         end_date: Union[str, date, None] = None):
    """Convert dataframe to real prices.

    Convert a dataframe's columns to real prices. Call the `cpi.get()`
    function to obtain the consumer price index. take into account the input
    dataframe's frequency and whether columns represent rolling averages or
    sums. Allow choosing a single period, a range of dates or no period as a
    base (i.e., period for which the average/sum of input dataframe and output
    dataframe is the same).

    Parameters
    ----------
    df : Pandas dataframe
    start_date : str, date or None (default is None)
        If set to a date-like string or a date, and `end_date` is None, the
        base period will be `start_date`.
    end_date : str, date or None (default is None)

    Returns
    -------
    converted_df : Pandas dataframe

    """
    inferred_freq = pd.infer_freq(df.index)

    cpi_data = cpi.get(update="cpi.csv", revise_rows=6, save="cpi.csv",
                       force_update=False)
    columns.set_metadata(cpi_data, ts_type="Flujo")
    cpi_freq = freqs.freq_resample(cpi_data, target=inferred_freq,
                                   operation="average").iloc[:, [0]]
    cum_periods = int(df.columns.get_level_values("Acum. períodos")[0])
    cpi_freq = freqs.rolling(cpi_freq, periods=cum_periods,
                             operation="average")
    cpi_to_use = cpi_freq[cpi_freq.index.isin(df.index)].iloc[:, 0]

    if start_date is None:
        converted_df = df.apply(lambda x:
                                x / cpi_to_use)
        col_text = "Const."
    elif end_date is None:
        converted_df = df.apply(lambda x:
                                x / cpi_to_use * cpi_to_use[start_date])
        col_text = f"Const. {start_date}"
    else:
        converted_df = df.apply(
            lambda x: x / cpi_to_use * cpi_to_use[start_date:end_date].mean()
        )
        col_text = f"Const. {start_date}_{end_date}"

    columns.set_metadata(converted_df, inf_adj=col_text)

    return converted_df


def pcgdp(df: pd.DataFrame, hifreq: bool = True):
    """Calculate dataframe as percentage of GDP.

    Convert a dataframe's columns to percentage of GDP. Call the
    `national_accounts.lin_gdp()` function to obtain UYU and USD quarterly GDP
    series. Take into account the input dataframe's currency for chossing UYU
    or USD GDP. If `hifreq` is set to `True`, GDP will be upsampled and linear
    interpolation will be performed to complete missing data.

    Parameters
    ----------
    df : Pandas dataframe
    hifreq : bool (default is True)
        If True, the input dataframe's frequency is assumed to be 'higher' than
        quarterly (`Q` or `Q-DEC`) and will trigger GDP upsampling.

    Returns
    -------
    converted_df : Pandas dataframe

    """
    inferred_freq = pd.infer_freq(df.index)
    gdp = national_accounts.lin_gdp(update="lin_gdp.csv",
                                    save="lin_gdp.csv", force_update=False)

    if hifreq is True:
        gdp = freqs.freq_resample(gdp, target=inferred_freq,
                                  operation="upsample", interpolation="linear")
    else:
        gdp = gdp.resample(inferred_freq, convention="end").asfreq()

    if df.columns.get_level_values("Unidad/Moneda")[0] == "USD":
        gdp = gdp.iloc[:, 1].to_frame()
    else:
        gdp = gdp.iloc[:, 0].to_frame()

    gdp_to_use = gdp[gdp.index.isin(df.index)].iloc[:, 0]
    converted_df = df.apply(lambda x: x / gdp_to_use).multiply(100)

    columns.set_metadata(converted_df, currency="% PBI")

    return converted_df



from typing import Optional

import pandas as pd


def set_metadata(
        df: pd.DataFrame, area: Optional[str] = None,
        currency: Optional[str] = None, inf_adj: Optional[str] = None,
        index: Optional[str] = None, seas_adj: Optional[str] = None,
        ts_type: Optional[str] = None, cumperiods: Optional[int] = None
):
    """Add a multiindex to a dataframe's columns.

    Characterize a dataframe by adding metadata to its column names by
    use of multiindexes.

    Parameters
    ----------
    df : Pandas dataframe
    area : str or None (default is None)
        Topic to which the data relates to.
    currency : str or None (default is None)
        Currency denomination.
    inf_adj : str or None (default is None)
        Whether the data is in constant prices.
    index : str or None (default is None)
        Whether the data is some type of index.
    seas_adj : str or None (default is None)
        Whether the data is seasonally adjusted.
    ts_type : str or None (default is None)
        Time series type, generally 'stock' or 'flow'.
    cumperiods : int or None (default is None)
        Number of periods accumulated per observation.

    Returns
    -------
    None

    See also
    --------
    Modifies the dataframe's column names in place.

    """
    colnames = df.columns
    inferred_freq = pd.infer_freq(df.index)
    if inferred_freq is None:
        print("Frequency could not be inferred from the index.")
        inferred_freq = "-"

    if not isinstance(df.columns, pd.MultiIndex):
        df.columns = pd.MultiIndex.from_product(
            [
                colnames, [area], [inferred_freq], [currency], [inf_adj],
                [index], [seas_adj], [ts_type], [cumperiods]
            ],
            names=[
                "Indicador", "Área", "Frecuencia", "Unidad/Moneda",
                "Inf. adj.", "Índice", "Seas. Adj.", "Tipo", "Acum. períodos"
            ]
        )
    else:
        df.columns = df.columns.set_levels([inferred_freq], level=2)

        if area is not None:
            df.columns = df.columns.set_levels([area], level=1)

        if currency is not None:
            df.columns = df.columns.set_levels([currency], level=3)

        if inf_adj is not None:
            df.columns = df.columns.set_levels([inf_adj], level=4)

        if index is not None:
            df.columns = df.columns.set_levels([index], level=5)

        if seas_adj is not None:
            df.columns = df.columns.set_levels([seas_adj], level=6)

        if ts_type is not None:
            df.columns = df.columns.set_levels([ts_type], level=7)

        if cumperiods is not None:
            df.columns = df.columns.set_levels([cumperiods], level=8)



import os
import datetime as dt
from typing import Union
from pathlib import Path

import pandas as pd


def check_modified(data_path: Union[str, Path]):
    """Check how long it has been since the file in path was modified."""
    modified_time = dt.datetime.fromtimestamp(os.path.getmtime(data_path))
    delta = (dt.datetime.now() - modified_time).days
    previous_data = pd.read_csv(data_path, sep=" ", index_col=0, header=[0, 1, 2, 3, 4, 5, 6, 7, 8])
    previous_data.index = pd.to_datetime(previous_data.index)

    return delta, previous_data


def revise(new_data: pd.DataFrame, prev_data: pd.DataFrame, revise_rows: int):
    """Replace n rows of data at the end of a dataframe with new data."""
    non_revised = prev_data[:len(prev_data)-revise_rows]
    revised = new_data[len(prev_data)-revise_rows:]
    non_revised.columns = new_data.columns
    updated = non_revised.append(revised, sort=False)

    return updated



import os
import datetime as dt
from pathlib import Path
from typing import Union

import pandas as pd
from pandas.tseries.offsets import MonthEnd

from econuy.config import ROOT_DIR
from econuy.processing import freqs, updates, columns, convert
from econuy.resources.utils import nat_accounts_metadata

DATA_PATH = os.path.join(ROOT_DIR, "data")
update_threshold = 80


def get(update: bool = False, revise_rows: int = 0,
        save: bool = False, force_update: bool = False):
    """Get national accounts data.

    Parameters
    ----------
    update : bool (default is False)
        If true, try to update existing data on disk.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save : bool (default is False)
        If true, save output dataframe in CSV format.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for national accounts data, 80 days)

    Returns
    -------
    parsed_excels : dictionary of Pandas dataframes
        Each dataframe corresponds to a national accounts table.

    """
    parsed_excels = {}
    for file, metadata in nat_accounts_metadata.items():

        if update is True:
            update_path = os.path.join(DATA_PATH, metadata['Name'] + ".csv")
            delta, previous_data = updates.check_modified(update_path)

            if delta < update_threshold and force_update is False:
                print(f"{metadata['Name']}.csv was modified within"
                      f" {update_threshold} day(s). Skipping download...")
                parsed_excels.update({metadata["Name"]: previous_data})
                continue

        raw = pd.read_excel(file, skiprows=9, nrows=metadata["Rows"])
        proc = (raw.drop(columns=["Unnamed: 0"]).
                       dropna(axis=0, how="all").dropna(axis=1, how="all"))
        proc = proc.transpose()
        proc.columns = metadata["Colnames"]
        proc.drop(["Unnamed: 1"], inplace=True)

        fix_na_dates(proc)

        if metadata["Index"] == "No":
            proc = proc.divide(1000)
        if update is True:
            proc = updates.revise(new_data=proc, prev_data=previous_data,
                                  revise_rows=revise_rows)
        proc = proc.apply(pd.to_numeric, errors="coerce")

        columns.set_metadata(proc, area="Actividad económica", currency="UYU",
                             inf_adj=metadata["Inf. Adj."],
                             index=metadata["Index"], seas_adj=metadata["Seas"],
                             ts_type="Flujo", cumperiods=1)

        if save is True:
            save_path = os.path.join(DATA_PATH, metadata['Name'] + ".csv")
            proc.to_csv(save_path, sep=" ")

        parsed_excels.update({metadata["Name"]: proc})

    return parsed_excels


def fix_na_dates(df):
    """Cleanup dates inplace in BCU national accounts files.

    Parameters
    ----------
    df : Pandas dataframe

    Returns
    -------
    None

    """
    df.index = df.index.str.replace("*", "")
    df.index = df.index.str.replace(r"\bI \b", "3-", regex=True)
    df.index = df.index.str.replace(r"\bII \b", "6-", regex=True)
    df.index = df.index.str.replace(r"\bIII \b", "9-", regex=True)
    df.index = df.index.str.replace(r"\bIV \b", "12-", regex=True)
    df.index = pd.to_datetime(df.index, format="%m-%Y") + MonthEnd(1)


def lin_gdp(update: Union[str, Path, None] = None,
            save: Union[str, Path, None] = None,
            force_update: bool = False):
    """Get nominal GDP data in UYU and USD with forecasts.

    Update nominal GDP data for use in the `convert.pcgdp()` function.
    Get IMF forecasts for year of last available data point and the next
    year (for example, if the last period available at the BCU website is
    september 2019, fetch forecasts for 2019 and 2020).

    Parameters
    ----------
    update : str, Path or None (default is None)
        Path or path-like string pointing to a CSV file for updating.
    save : str, Path or None (default is None)
        Path or path-like string where to save the output dataframe in CSV
        format.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for national accounts data, 80 days)

    Returns
    -------
    output : Pandas dataframe
        Quarterly GDP in UYU and USD with 1 year forecasts.

    """
    if update is not None:
        update_path = os.path.join(DATA_PATH, update)
        delta, previous_data = updates.check_modified(update_path)

        if delta < update_threshold and force_update is False:
            print(f"{update} was modified within {update_threshold} day(s). "
                  f"Skipping download...")
            return previous_data

    data_uyu = get(update=True, revise_rows=4, save=True,
                   force_update=False)["na_gdp_cur_nsa"]
    data_uyu = freqs.rolling(data_uyu, periods=4, operation="sum")
    data_usd = convert.usd(data_uyu)

    data = [data_uyu, data_usd]
    last_year = data_uyu.index.max().year
    if data_uyu.index.max().month == 12:
        last_year += 1

    results = []
    for table, gdp in zip(["NGDP", "NGDPD"], data):

        table_url = (f"https://www.imf.org/external/pubs/ft/weo/2019/02/weodat"
                     f"a/weorept.aspx?sy={last_year-1}&ey={last_year+1}&scsm=1"
                     f"&ssd=1&sort=country&ds=.&br=1&pr1.x=27&pr1.y=9&c=298&s"
                     f"={table}&grp=0&a=")
        imf_data = pd.to_numeric(pd.read_html(table_url)[4].iloc[2, [5, 6, 7]])
        imf_data = imf_data.reset_index(drop=True)
        fcast = (gdp.loc[[dt.datetime(last_year-1, 12, 31)]].
                 multiply(imf_data.iloc[1]).divide(imf_data.iloc[0]))
        fcast = fcast.rename(index={dt.datetime(last_year-1, 12, 31):
                                    dt.datetime(last_year, 12, 31)})
        next_fcast = (gdp.loc[[dt.datetime(last_year-1, 12, 31)]].
                      multiply(imf_data.iloc[2]).divide(imf_data.iloc[0]))
        next_fcast = next_fcast.rename(index={dt.datetime(last_year-1, 12, 31):
                                              dt.datetime(last_year+1, 12, 31)})
        fcast = fcast.append(next_fcast)
        gdp = gdp.append(fcast)
        results.append(gdp)

    output = pd.concat(results, axis=1)
    output = output.resample("Q-DEC").interpolate("linear")

    if save is not None:
        save_path = os.path.join(DATA_PATH, save)
        output.to_csv(save_path, sep=" ")

    return output



import os

import pandas as pd

from econuy.config import ROOT_DIR
from econuy.processing import columns
from econuy.retrieval import fx_ff, reserves_chg

data_path = os.path.join(ROOT_DIR, "data")
path_chg = os.path.join(data_path, "reserves_chg.csv")
path_ff = os.path.join(data_path, "fx_ff.csv")
path_save = os.path.join(data_path, "fx_spot_ff.csv")

changes = reserves_chg.get_reserves_chg(files=reserves_chg.files_,
                                        online_files=None, offline_files=None,
                                        update=path_chg, save=path_chg)
ff = fx_ff.get(update=path_ff, save=path_ff)

spot = changes.iloc[:, 0]

fx_ops = pd.merge(spot, ff, how="outer", left_index=True, right_index=True)
fx_ops = fx_ops.loc[(fx_ops.index >= ff.index.min()) &
                    (fx_ops.index <= spot.index.max())]
fx_ops = fx_ops.apply(pd.to_numeric, errors="coerce")
fx_ops = fx_ops.fillna(0)
fx_ops.columns = ["Spot", "Futuros", "Forwards"]

columns.set_metadata(fx_ops, area="Reservas internacionales", currency="USD",
                     inf_adj="No", index="No", seas_adj="NSA", ts_type="Flujo",
                     cumperiods=1)

fx_ops.to_csv(path_save, sep=" ")



import datetime as dt
import os
import urllib
from pathlib import Path
from typing import List, Union

import pandas as pd
from dateutil.relativedelta import relativedelta

from econuy.config import ROOT_DIR
from econuy.processing import columns
from econuy.resources.utils import reserves_cols, reserves_url

months = ["ene", "feb", "mar", "abr", "may", "jun",
          "jul", "ago", "set", "oct", "nov", "dic"]
years = list(range(2013, dt.datetime.now().year + 1))
files_ = [month + str(year) for year in years for month in months]

DATA_PATH = os.path.join(ROOT_DIR, "data")


def base_reports(files: List[str], update: Union[str, Path, None] = None):
    """Get international reserves change data from online sources.

    Use as input a list of strings of the format %b%Y, each representing a
    month of data.

    Parameters
    ----------
    files : list of strings
        List of strings of the type '%b%Y', only %b is in Spanish. So 'ene'
        instead of 'jan'. For example, 'oct2017'.
    update : str, Path or None (default is None)
        Path or path-like string pointing to a CSV file for updating.

    Returns
    -------
    reserves : Pandas dataframe

    """
    urls = [f"{reserves_url}{file}.xls" for file in files]

    if update is not None:
        update_path = os.path.join(DATA_PATH, update)
        urls = urls[-18:]
        previous_data = pd.read_csv(update_path, sep=" ", index_col=0,
                                    header=list(range(9)))
        previous_data.columns = reserves_cols[1:46]
        previous_data.index = pd.to_datetime(previous_data.index)

    reports = []
    for url in urls:

        try:
            with pd.ExcelFile(url) as xls:
                month_of_report = pd.read_excel(xls, sheet_name="INDICE")
                raw = pd.read_excel(xls, sheet_name="ACTIVOS DE RESERVA",
                                    skiprows=3)

            first_day = month_of_report.iloc[7, 4]
            last_day = (first_day
                        + relativedelta(months=1)
                        - dt.timedelta(days=1))

            proc = raw.dropna(axis=0, thresh=20).dropna(axis=1, thresh=20)
            proc = proc.transpose()
            proc.index.name = "Date"
            proc = proc.iloc[:, 1:46]
            proc.columns = reserves_cols[1:46]
            proc = proc.iloc[1:]
            proc.index = pd.to_datetime(proc.index, errors="coerce")
            proc = proc.loc[proc.index.dropna()]
            proc = proc.loc[first_day:last_day]

            reports.append(proc)

        except urllib.error.HTTPError:
            print(f"{url} could not be reached.")
            pass

    reserves = pd.concat(reports, sort=False)

    if update is not None:
        reserves = previous_data.append(reserves, sort=False)
        reserves = reserves.loc[~reserves.index.duplicated(keep="last")]

    reserves = reserves.apply(pd.to_numeric, errors="coerce")

    return reserves


def missing_reports(online_files: List[str], offline_files: List[str]):
    """Get missing reserves data from online and offline sources.

    Parameters
    ----------
    online_files : list of strings
        Used for reserves files that do not match the '%b%Y' format. See
        `base_reports`.
    offline_files : list of strings
        List of strings pointing to CSV locations.

    Returns
    -------
    missing : Pandas dataframe

    """
    missing_online = base_reports(online_files)

    missing_offline = []
    for file in offline_files:
        offline_aux = pd.read_csv(file, sep=" ", index_col=0)
        offline_aux.index = pd.to_datetime(offline_aux.index, errors="coerce")

        missing_offline.append(offline_aux)

    missing_offline = pd.concat(missing_offline, sort=False)

    missing = missing_online.append(missing_offline, sort=False)
    missing = missing.apply(pd.to_numeric, errors="coerce")

    return missing


def get_reserves_chg(files: List[str], online_files: List[str] = None,
                     offline_files: List[str] = None,
                     update: Union[str, Path, None] = None,
                     save: Union[str, Path, None] = None):
    """Get international reserves changes data from online and offline sources.

    Call the `base_reports()` and `missing_reports()` functions.

    Parameters
    ----------
    files : list of strings
    online_files : list of strings (default is None)
        Used for reserves files that do not match the '%b%Y' format. See
        `base_reports`.
    offline_files : list of strings (default is None)
        List of strings pointing to CSV locations.
    update : str, Path or None (default is None)
        Path or path-like string pointing to a CSV file for updating.
    save : str, Path or None (default is None)
        Path or path-like string where to save the output dataframe in CSV
        format.

    Returns
    -------
    reserves : Pandas dataframe

    """
    reserves = base_reports(files=files, update=update)

    if online_files is not None or offline_files is not None:
        missing = missing_reports(online_files=online_files,
                                  offline_files=offline_files)
        reserves = reserves.append(missing, sort=False)
        reserves.sort_index(inplace=True)

    columns.set_metadata(reserves, area="Reservas internacionales",
                         currency="USD", inf_adj="No", index="No",
                         seas_adj="NSA", ts_type="Flujo", cumperiods=1)

    if save is not None:
        save_path = os.path.join(DATA_PATH, save)
        reserves.to_csv(save_path, sep=" ")

    return reserves



import os
from pathlib import Path
from typing import Union

import pandas as pd

from econuy.config import ROOT_DIR
from econuy.processing import updates, columns
from econuy.resources.utils import labor_url

DATA_PATH = os.path.join(ROOT_DIR, "data")
update_threshold = 25


def get(update: Union[str, Path, None] = None, revise_rows: int = 0,
        save: Union[str, Path, None] = None, force_update: bool = False):
    """Get labor market data.

    Get monthly labor force participation rate, employment rate (employment to
    working-age population) and unemployment rate.

    Parameters
    ----------
    update : str, Path or None (default is None)
        Path or path-like string pointing to a CSV file for updating.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save : str, Path or None (default is None)
        Path or path-like string where to save the output dataframe in CSV
        format.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for labor market, 25 days)

    Returns
    -------
    labor : Pandas dataframe

    """
    if update is not None:
        update_path = os.path.join(DATA_PATH, update)
        delta, previous_data = updates.check_modified(update_path)

        if delta < update_threshold and force_update is False:
            print(f"{update} was modified within {update_threshold} day(s). "
                  f"Skipping download...")
            return previous_data

    labor_raw = pd.read_excel(labor_url, skiprows=39).dropna(axis=0, thresh=2)
    labor = labor_raw[~labor_raw["Unnamed: 0"].str.contains("-|/|Total",
                                                            regex=True)]
    labor = labor[["Unnamed: 1", "Unnamed: 4", "Unnamed: 7"]]
    labor.index = pd.date_range(start="2006-01-01",
                                periods=len(labor), freq="M")
    labor.columns = ["Tasa de actividad", "Tasa de empleo",
                     "Tasa de desempleo"]

    if update is not None:
        labor = updates.revise(new_data=labor, prev_data=previous_data,
                               revise_rows=revise_rows)

    labor = labor.apply(pd.to_numeric, errors="coerce")
    columns.set_metadata(labor, area="Mercado laboral", currency="-",
                         inf_adj="No", index="No", seas_adj="NSA",
                         ts_type="-", cumperiods=1)

    if save is not None:
        save_path = os.path.join(DATA_PATH, save)
        labor.to_csv(save_path, sep=" ")

    return labor



import os
import datetime as dt
import zipfile
import tempfile
from io import BytesIO
from pathlib import Path
from typing import Union

import pandas as pd
import numpy as np
from pandas.tseries.offsets import YearEnd, MonthEnd
import requests

from econuy.config import ROOT_DIR
from econuy.processing import updates
from econuy.resources.utils import (beef_url, pulp_url, soybean_url,
                                    what_url, imf_url, milk1_url, milk2_url)

DATA_PATH = os.path.join(ROOT_DIR, "data")
UPDATE_THRESHOLD = 10


def weights(window: int = 3):
    """Prepare weights for the commodity price index.

    Parameters
    ----------
    window : int (default is 3)
        How many years to consider for rolling averages.

    Returns
    -------
    resampled_weights : Pandas dataframe
        Export-based weights for relevant commodities to Uruguay.

    """
    weights_path = os.path.join(DATA_PATH, "other",
                                "commodity_exports_wits.csv")

    raw = pd.read_csv(weights_path)
    table = raw.pivot(index="Year", columns="ProductDescription",
                      values="TradeValue in 1000 USD")
    table.fillna(0, inplace=True)
    not_considered = ["Maize except sweet corn.", "Wine of fresh grapes",
                      "Hide/skin/fur, raw", "Sugar/sugar prep/honey",
                      "Fish/shellfish/etc.", "Vegetables and fruit",
                      "Malt/ malt flour"]
    table.drop(not_considered, axis=1, inplace=True)
    percentage = table.div(table.sum(axis=1), axis=0)
    percentage.index = (pd.to_datetime(percentage.index, format="%Y")
                        + YearEnd(1))
    rolling = percentage.rolling(window=window, min_periods=window).mean()
    resampled_weights = rolling.resample("M").bfill()

    beef = ["Beef offal, frozen", "Beef offal,fresh/chilled",
            "Beef prepared/presvd nes", "Beef, fresh/chilld/frozn",
            "Bovine animals, live"]
    resampled_weights["Beef"] = resampled_weights[beef].sum(
        axis=1, min_count=len(beef)
    )
    resampled_weights.drop(beef, axis=1, inplace=True)
    resampled_weights.columns = ["Barley", "Wood", "Gold", "Milk", "Pulp",
                                 "Rice", "Soybeans", "Wheat", "Wool", "Beef"]

    return resampled_weights


def prices(update: Union[str, Path, None] = None, revise_rows: int = 0,
           save: Union[str, Path, None] = None, force_update: bool = False):
    """Prepare prices for the commodity price index.

    Parameters
    ----------
    update : str, Path or None (default is None)
        Path or path-like string pointing to a CSV file for updating.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save : str, Path or None (default is None)
        Path or path-like string where to save the output dataframe in CSV
        format.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for commodity prices, 10 days)

    Returns
    -------
    complete : Pandas dataframe
        Prices and price indexes of relevant commodities for Uruguay.

    """
    bushel_conv = 36.74 / 100

    if update is not None:
        update_path = os.path.join(DATA_PATH, update)
        modified_time = dt.datetime.fromtimestamp(os.path.getmtime(update_path))
        delta = (dt.datetime.now() - modified_time).days
        previous_data = pd.read_csv(update_path, sep=" ", index_col=0)
        previous_data.index = pd.to_datetime(previous_data.index)

        if delta < UPDATE_THRESHOLD and force_update is False:
            print(f"{update} was modified within {UPDATE_THRESHOLD} day(s). "
                  f"Skipping download...")
            return previous_data

    raw_beef = (pd.read_excel(beef_url, header=4, index_col=0)
                .dropna(how="all").drop(index=0))
    raw_beef.columns = raw_beef.columns.str.strip()
    proc_beef = raw_beef["Ing. Prom./Ton."].to_frame()
    proc_beef.index = pd.date_range(start="2002-01-04",
                                    periods=len(proc_beef), freq="W-SAT")
    proc_beef["Ing. Prom./Ton."] = np.where(
        proc_beef > np.mean(proc_beef) + np.std(proc_beef) * 2,
        proc_beef / 1000,
        proc_beef,
    )
    beef = proc_beef.resample("M").mean()

    raw_pulp_r = requests.get(pulp_url)
    temp_dir = tempfile.TemporaryDirectory()
    with zipfile.ZipFile(BytesIO(raw_pulp_r.content), "r") as f:
        f.extractall(path=temp_dir.name)
        path = os.path.join(temp_dir.name, "monthly_values.csv")
        raw_pulp = pd.read_csv(path, sep=";").dropna(how="any")
    proc_pulp = raw_pulp.copy().sort_index(ascending=False)
    proc_pulp.index = pd.date_range(start="1990-01-31",
                                    periods=len(proc_pulp), freq="M")
    proc_pulp.drop(["Label", "Codes"], axis=1, inplace=True)
    pulp = proc_pulp

    soy_wheat = []
    for url in [soybean_url, what_url]:
        raw = pd.read_csv(url, index_col=0)
        proc = (raw["Settle"] * bushel_conv).to_frame()
        proc.index = pd.to_datetime(proc.index, format="%Y-%m-%d")
        proc.sort_index(inplace=True)
        soy_wheat.append(proc.resample("M").mean())
    soybean = soy_wheat[0]
    wheat = soy_wheat[1]

    milk = []
    for region, row in {milk1_url: 14, milk2_url: 13}.items():
        raw_milk = pd.read_excel(region, skiprows=row,
                                 nrows=dt.datetime.now().year - 2007)
        raw_milk.dropna(how="all", axis=1, inplace=True)
        raw_milk.drop(["Promedio ", "Variación"], axis=1, inplace=True)
        raw_milk.columns = ["Año/Mes"] + list(range(1, 13))
        proc_milk = pd.melt(raw_milk, id_vars=["Año/Mes"])
        proc_milk.sort_values(by=["Año/Mes", "variable"], inplace=True)
        proc_milk.index = pd.date_range(start="2007-01-31",
                                        periods=len(proc_milk), freq="M")
        milk.append(proc_milk.iloc[:, 2].to_frame())
    milk = pd.concat(milk, axis=1).mean(axis=1).to_frame()
    milk.columns = ["Price"]
    prev_milk = pd.read_csv(os.path.join(DATA_PATH, "other",
                                         "milk_oceania_europe_fao.csv"),
                            index_col=0)
    prev_milk.index = pd.to_datetime(prev_milk.index,
                                     format="%b-%y") + MonthEnd(1)
    prev_milk = prev_milk.replace(",", "", regex=True)
    prev_milk = pd.to_numeric(prev_milk.squeeze()).to_frame()
    prev_milk = prev_milk.loc[prev_milk.index < min(milk.index)]
    prev_milk.columns = ["Price"]
    milk = prev_milk.append(milk)

    raw_imf = pd.read_excel(imf_url)
    raw_imf.columns = raw_imf.iloc[0, :]
    proc_imf = raw_imf.iloc[3:, 1:]
    proc_imf.index = pd.date_range(start="1980-01-31",
                                   periods=len(proc_imf), freq="M")
    rice = proc_imf[proc_imf.columns[proc_imf.columns.str.contains("Rice")]]
    wood = proc_imf[proc_imf.columns[
        proc_imf.columns.str.contains("Sawnwood")
    ]]
    wood = wood.mean(axis=1).to_frame()
    wool = proc_imf[proc_imf.columns[proc_imf.columns.str.startswith("Wool")]]
    wool = wool.mean(axis=1).to_frame()
    barley = proc_imf[proc_imf.columns[
        proc_imf.columns.str.startswith("Barley")
    ]]
    gold = proc_imf[proc_imf.columns[
        proc_imf.columns.str.startswith("Gold")
    ]]

    complete = pd.concat([beef, pulp, soybean, milk, rice, wood, wool, barley,
                          gold, wheat], axis=1)
    complete = complete.reindex(beef.index).dropna(thresh=9)
    complete.columns = ["Beef", "Pulp", "Soybeans", "Milk", "Rice", "Wood",
                        "Wool", "Barley", "Gold", "Wheat"]

    if update is not None:
        complete = updates.revise(new_data=complete, prev_data=previous_data,
                                  revise_rows=revise_rows)

    if save is not None:
        save_path = os.path.join(DATA_PATH, save)
        complete.to_csv(save_path, sep=" ")

    return complete


def get(save: Union[str, Path, None] = None):
    """Get the commodity price index.

    Parameters
    ----------
    save : str, Path or None (default is None)
        Path or path-like string where to save the output dataframe in CSV
        format.

    Returns
    -------
    product : Pandas dataframe
        Export-weighted average of commodity prices relevant to Uruguay.

    """
    prices_ = prices(update="commodity_prices.csv", revise_rows=3,
                     save="commodity_prices.csv", force_update=False)
    prices_ = prices_.interpolate(method="linear", limit=1).dropna(how="any")
    prices_ = prices_.pct_change(periods=1)
    weights_ = weights(window=3)
    weights_ = weights_[prices_.columns]
    weights_ = weights_.reindex(prices_.index, method="ffill")

    product = pd.DataFrame(prices_.values * weights_.values,
                           columns=prices_.columns, index=prices_.index)
    product = product.sum(axis=1).add(1).to_frame().cumprod()

    if save is not None:
        save_path = os.path.join(DATA_PATH, save)
        product.to_csv(save_path, sep=" ")

    return product



from datetime import date
from typing import Union

import pandas as pd

from econuy.processing import columns


def base_index(df: pd.DataFrame, start_date: Union[str, date],
               end_date: Union[str, date, None] = None, base: float = 100):
    """Rebase all dataframe columns to a date or range of dates.

    Parameters
    ----------
    df : Pandas dataframe
    start_date : string or date
        Date to which series will be rebased.
    end_date : string or date (default is `None`)
        If specified, series will be rebased to the average between start_date
        and end_date.
    base : float (default is 100)
        Float for which start_date == float or average between start_date and
        end_date == float.

    Returns
    -------
    indexed : Pandas dataframe

    """
    if end_date is None:
        indexed = df.apply(lambda x: x / x[start_date] * base)
        columns.set_metadata(indexed, index=start_date)

    else:
        indexed = df.apply(lambda x: x / x[start_date:end_date].mean() * base)
        columns.set_metadata(indexed, index=f"{start_date}_{end_date}")

    return indexed



from typing import Optional

import pandas as pd

from econuy.processing import columns

PD_FREQUENCIES = {"A": 1,
                  "Q-DEC": 4,
                  "M": 12}


def freq_resample(df: pd.DataFrame, target: str, operation: str = "sum",
                  interpolation: str = "linear"):
    """Wrapper for the resample method in Pandas.

    Resample taking into account dataframe `type` so that stock data is not
    averaged or summed when resampling; only last value of target frequency
    is considered.

    Parameters
    ----------
    df : Pandas dataframe
    target : str
        Target frequency to resample to. See
        https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
    operation : str (default is 'sum')
        'sum', 'average' and 'upsample' are allowed.
    interpolation : bool (default is True)
        Method to use when missing data are produced as a result of resampling.
        See https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.interpolate.html

    Returns
    -------
    resampled_df : Pandas dataframe

    Raises
    ------
    ValueError:
        If `operation` is not one of available options and if the input
        dataframe does not have a Type level in its column multiindex.

    """
    if df.columns.get_level_values("Tipo")[0] == "Flujo":
        if operation == "sum":
            resampled_df = df.resample(target).sum()
        elif operation == "average":
            resampled_df = df.resample(target).mean()
        elif operation == "upsample":
            resampled_df = df.resample(target).asfreq()
            resampled_df = resampled_df.interpolate(method=interpolation)
        else:
            raise ValueError("Only sum, average and upsample "
                             "are accepted operations")

        cum_periods = int(df.columns.get_level_values("Acum. períodos")[0])
        if cum_periods != 1:
            input_notna = df.iloc[:, 0].count()
            output_notna = resampled_df.iloc[:, 0].count()
            cum_adj = round(output_notna / input_notna)
            columns.set_metadata(resampled_df,
                                 cumperiods=int(cum_periods * cum_adj))

    elif df.columns.get_level_values("Tipo")[0] == "Stock":
        resampled_df = df.resample(target, convention="end").asfreq()
        resampled_df = resampled_df.interpolate(method=interpolation)
    else:
        raise ValueError("Dataframe needs to have a Type")

    columns.set_metadata(resampled_df)

    return resampled_df


def rolling(df: pd.DataFrame, periods: Optional[int] = None,
            operation: str = "sum"):
    """Wrapper for the rolling method in Pandas.

    If `periods` is None, try to infer the frequency and set `periods`
    according to the following logic: `{'A': 1, 'Q-DEC': 4, 'M': 12}`.

    Parameters
    ----------
    df : Pandas dataframe
    periods : Optional[int] (default is None)
    operation : str (default is 'sum')
        'sum' and 'average' are allowed.

    Returns
    -------
    rolling_df : Pandas dataframe

    """
    window_operation = {
        "sum": lambda x: x.rolling(window=periods,
                                   min_periods=periods).sum(),
        "average": lambda x: x.rolling(window=periods,
                                       min_periods=periods).mean()
    }

    if df.columns.get_level_values("Tipo")[0] == "Stock":
        raise Warning("Rolling operations shouldn't be "
                      "calculated on stock variables")

    if periods is None:
        inferred_freq = pd.infer_freq(df.index)
        periods = PD_FREQUENCIES[inferred_freq]

    rolling_df = df.apply(window_operation[operation])

    columns.set_metadata(rolling_df, cumperiods=periods)

    return rolling_df



import datetime as dt
import os
import urllib
from pathlib import Path
from typing import Union

import pandas as pd

from econuy.config import ROOT_DIR
from econuy.processing import columns
from econuy.resources.utils import ff_url

DATA_PATH = os.path.join(ROOT_DIR, "data")


def get(update: Union[str, Path, None] = None,
        save: Union[str, Path, None] = None):
    """Get future and forwards FX operations by the Central Bank.

    Parameters
    ----------
    update : str, Path or None (default is None)
        Path or path-like string pointing to a CSV file for updating.
    save : str, Path or None (default is None)
        Path or path-like string where to save the output dataframe in CSV
        format.

    Returns
    -------
    operations : Pandas dataframe
        Daily future and forward FX operations since november 2013.

    """
    dates = pd.bdate_range("2013-11-01",
                           dt.datetime.today()).strftime("%y%m%d").tolist()
    if update is not None:
        update_path = os.path.join(DATA_PATH, update)
        prev_data = pd.read_csv(update_path, sep=" ", index_col=0,
                                header=[0, 1, 2, 3, 4, 5, 6, 7, 8])
        prev_data.columns = ["Futuros", "Forwards"]
        prev_data.index = pd.to_datetime(prev_data.index)
        last_date = prev_data.index[len(prev_data)-1]
        dates = pd.bdate_range(last_date,
                               dt.datetime.today()).strftime("%y%m%d").tolist()

    reports = []
    for date in dates:

        try:
            raw_report = pd.read_excel(f"{ff_url}{date}.xls")

            if (dt.datetime.strptime(date, "%y%m%d") >=
                    dt.datetime(2014, 5, 21)):
                future1 = raw_report.iloc[19, ].apply(pd.to_numeric,
                                                      errors="coerce").sum()
                future2 = raw_report.iloc[21, ].apply(pd.to_numeric,
                                                      errors="coerce").sum()
                future = future1 + future2
                forward = raw_report.iloc[24, ].apply(pd.to_numeric,
                                                      errors="coerce").sum()

            else:
                future = raw_report.iloc[19, ].apply(pd.to_numeric,
                                                     errors="coerce").sum()
                forward = raw_report.iloc[22, ].apply(pd.to_numeric,
                                                      errors="coerce").sum()

            reports.append([dt.datetime.strptime(date, "%y%m%d"),
                            future, forward])

        except urllib.error.HTTPError:
            print(f"Report for {date} could not be reached.")
            pass

    operations = pd.DataFrame(reports)
    operations.columns = ["Date", "Futuros", "Forwards"]
    operations.set_index("Date", inplace=True)
    operations = operations.divide(1000)

    if update is not None:
        operations = prev_data.append(operations, sort=False)

    columns.set_metadata(
        operations, area="Reservas internacionales",  currency="USD",
        inf_adj="No", index="No", seas_adj="NSA", ts_type="Flujo",
        cumperiods=1)

    if save is not None:
        save_path = os.path.join(DATA_PATH, save)
        operations.to_csv(save_path, sep=" ")

    return operations



import os
from setuptools import setup, find_packages
import json


def get_requirements_from_pipfile_lock(pipfile_lock=None):
    if pipfile_lock is None:
        pipfile_lock = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), 'Pipfile.lock'
        )
    lock_data = json.load(open(pipfile_lock))
    return [package_name for package_name in
            lock_data.get('default', {}).keys()]


packages = find_packages('.', exclude=['*.test', '*.test.*'])
pipfile_lock_requirements = get_requirements_from_pipfile_lock()

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="econuy",
    version="0.1.1",
    description="Download and process Uruguayan economic data.",
    long_description=long_description,
    author="Rafael Xavier",
    license="GPL-3.0",
    url="https://github.com/rxavier/uy-econ",
    keywords=["uruguay", "economy", "economic", "statistics", "data"],
    install_requires=pipfile_lock_requirements,
    packages=packages,
    python_requires=">=3.7"
)



import os
from datetime import date
from typing import Union

import pandas as pd

from econuy.config import ROOT_DIR
from econuy.retrieval import nxr, national_accounts, cpi, fiscal_accounts, \
    labor
from econuy.processing import columns, freqs, variations, seasonal, convert
from econuy.resources.utils import fiscal_metadata

DATA_PATH = os.path.join(ROOT_DIR, "data")


def inflation():
    """Update CPI data and return common inflation measures.

    Returns
    -------
    dataframe : Pandas dataframe
        Columns: CPI index, annual inflation, monthly inflation, seasonally
        adjusted monthly inflation and trend monthly inflation.

    """
    data = cpi.get(update="cpi.csv", revise_rows=6,
                   save="cpi.csv", force_update=False)
    interannual = variations.chg_diff(data, operation="chg", period_op="inter")
    monthly = variations.chg_diff(data, operation="chg", period_op="last")
    trend, seasadj = seasonal.decompose(data, trading=True, outlier=False)
    monthly_sa = variations.chg_diff(seasadj, operation="chg",
                                     period_op="last")
    monthly_trend = variations.chg_diff(trend, operation="chg",
                                        period_op="last")

    output = pd.concat([data, interannual, monthly,
                        monthly_sa, monthly_trend], axis=1)

    return output


def exchange_rate(eop: bool = False, sell: bool = True,
                  seas_adj: Union[str, None] = None, cum: int = 1):
    """Get nominal exchange rate data.

    Allow choosing end of period or average (monthly), sell/buy rates,
    calculating seasonal decomposition and rolling averages.

    Parameters
    ----------
    eop : bool (default is False)
        End of period data.
    sell : bool (default is True)
        Sellr ate.
    seas_adj : str or None (default is None)
        Allowed strings: 'trend' or 'seas'.
    cum : int (default is 1)
        How many periods to accumulate for rolling averages.

    Returns
    -------
    dataframe : Pandas dataframe

    """
    data = nxr.get(update="nxr.csv", revise_rows=6,
                   save="nxr.csv", force_update=False)

    if eop is False:
        output = data.iloc[:, [2, 3]]
    else:
        output = data.iloc[:, [0, 1]]

    if sell is True:
        output = output.iloc[:, 1].to_frame()
    else:
        output = output.iloc[:, 0].to_frame()

    if seas_adj in ["trend", "seas"] and cum == 1:
        trend, seasadj = seasonal.decompose(output, trading=True, outlier=True)
        if seas_adj == "trend":
            output = pd.concat([output, trend], axis=1)
        elif seas_adj == "seas":
            output = pd.concat([output, seasadj], axis=1)
        else:
            print("Only 'trend', 'seas' and None are "
                  "possible values for 'seas_adj'")

    if cum != 1:
        output = freqs.rolling(output, periods=cum, operation="average")

    return output


def fiscal(aggregation: str = "gps", fss: bool = True,
           unit: Union[str, None] = "gdp",
           start_date: Union[str, date, None] = None,
           end_date: Union[str, date, None] = None, cum: int = 1,
           seas_adj: Union[str, None] = None):
    """Get fiscal accounts data.

    Allow choosing government aggregation, whether to exclude the FSS
    (Fideicomiso  de la Seguridad Social, Social Security Trust Fund), the unit
    (UYU, real UYU, USD, real USD or percent of GDP), periods to accumuldate
    for rolling sums and seasonal adjustment.

    Parameters
    ----------
    aggregation : str (default is 'gps')
        Government aggregation. Can be 'gps' (consolidated public sector),
        'nfps' (non-financial public sector) or 'gc' (central government).
    fss : bool (default is True)
        If True, exclude the FSS's income from gov't revenues and the FSS's
        interest revenues from gov't interest payments.
    unit : str or None (default is 'gdp')
        Unit in which data should be expressed. Possible values are 'real',
        'usd', 'real usd' and 'gdp'. If None or another string is set, no unit
        calculations will be performed, rendering the data as is (current UYU).
    start_date : str, date or None (default is None)
        If `unit` is set to 'real' or 'real usd', this parameter and `end_date`
        control how deflation is calculated.
    end_date :
        If `unit` is set to 'real' or 'real usd', this parameter and
        `start_date` control how deflation is calculated.
    cum : int (default is 1)
        How many periods to accumulate for rolling sums.
    seas_adj :
        Allowed strings: 'trend' or 'seas'.

    Returns
    -------
    dataframe : Pandas dataframe

    """
    data = fiscal_accounts.get(update=True, revise_rows=12,
                               save=True, force_update=False)
    gps = data["fiscal_gps"]
    nfps = data["fiscal_nfps"]
    gc = data["fiscal_gc-bps"]

    proc = pd.DataFrame(index=gps.index)
    proc["Ingresos: SPNF-SPC"] = nfps["Ingresos: SPNF"]
    proc["Ingresos: GC-BPS"] = gc["Ingresos: GC-BPS"]
    proc["Egresos: Primarios SPNF-SPC"] = nfps["Egresos: Primarios SPNF"]
    proc["Egresos: Totales GC-BPS"] = gc["Egresos: GC-BPS"]
    proc["Egresos: Inversiones SPNF-SPC"] = nfps["Egresos: Inversiones"]
    proc["Egresos: Inversiones GC-BPS"] = gc["Egresos: Inversión"]
    proc["Intereses: SPNF"] = nfps["Intereses: Totales"]
    proc["Intereses: BCU"] = gps["Intereses: BCU"]
    proc["Intereses: SPC"] = proc["Intereses: SPNF"] + proc["Intereses: BCU"]
    proc["Intereses: GC-BPS"] = gc["Intereses: Total"]
    proc["Egresos: Totales SPNF"] = (proc["Egresos: Primarios SPNF-SPC"]
                                     + proc["Intereses: SPNF"])
    proc["Egresos: Totales SPC"] = (proc["Egresos: Totales SPNF"]
                                    + proc["Intereses: BCU"])
    proc["Egresos: Primarios GC-BPS"] = (proc["Egresos: Totales GC-BPS"]
                                        - proc["Intereses: GC-BPS"])
    proc["Resultado: Primario intendencias"] = nfps[
        "Resultado: Primario intendencias"
    ]
    proc["Resultado: Primario BSE"] = nfps["Resultado: Primario BSE"]
    proc["Resultado: Primario BCU"] = gps["Resultado: Primario BCU"]
    proc["Resultado: Primario SPNF"] = nfps["Resultado: Primario SPNF"]
    proc["Resultado: Global SPNF"] = nfps["Resultado: Global SPNF"]
    proc["Resultado: Primario SPC"] = gps["Resultado: Primario SPC"]
    proc["Resultado: Global SPC"] = gps["Resultado: Global SPC"]
    proc["Resultado: Primario GC-BPS"] = (proc["Ingresos: GC-BPS"]
                                          - proc["Egresos: Primarios GC-BPS"])
    proc["Resultado: Global GC-BPS"] = gc["Resultado: Global GC-BPS"]

    proc["Ingresos: FSS"] = gc["Ingresos: FSS"]
    proc["Intereses: FSS"] = gc["Intereses: BPS-FSS"]
    proc["Ingresos: SPNF-SPC aj. FSS"] = (proc["Ingresos: SPNF-SPC"]
                                          - proc["Ingresos: FSS"])
    proc["Ingresos: GC-BPS aj. FSS"] = (proc["Ingresos: GC-BPS"]
                                        - proc["Ingresos: FSS"])
    proc["Intereses: SPNF aj. FSS"] = (proc["Intereses: SPNF"]
                                       - proc["Intereses: FSS"])
    proc["Intereses: SPC aj. FSS"] = (proc["Intereses: SPC"]
                                      - proc["Intereses: FSS"])
    proc["Intereses: GC-BPS aj. FSS"] = (proc["Intereses: GC-BPS"]
                                         - proc["Intereses: FSS"])
    proc["Egresos: Totales SPNF aj. FSS"] = (proc["Egresos: Totales SPNF"]
                                             - proc["Intereses: FSS"])
    proc["Egresos: Totales SPC aj. FSS"] = (proc["Egresos: Totales SPC"]
                                            - proc["Intereses: FSS"])
    proc["Egresos: Totales GC-BPS aj. FSS"] = (proc["Egresos: Totales GC-BPS"]
                                               - proc["Intereses: FSS"])
    proc["Resultado: Primario SPNF aj. FSS"] = (proc["Resultado: Primario SPNF"]
                                                - proc["Ingresos: FSS"])
    proc["Resultado: Global SPNF aj. FSS"] = (proc["Resultado: Global SPNF"]
                                              - proc["Ingresos: FSS"]
                                              + proc["Intereses: FSS"])
    proc["Resultado: Primario SPC aj. FSS"] = (proc["Resultado: Primario SPC"]
                                               - proc["Ingresos: FSS"])
    proc["Resultado: Global SPC aj. FSS"] = (proc["Resultado: Global SPC"]
                                             - proc["Ingresos: FSS"]
                                             + proc["Intereses: FSS"])
    proc["Resultado: Primario GC-BPS aj. FSS"] = (proc["Resultado: Primario GC-BPS"]
                                                  - proc["Ingresos: FSS"])
    proc["Resultado: Global GC-BPS aj. FSS"] = (proc["Resultado: Global GC-BPS"]
                                                - proc["Ingresos: FSS"]
                                                + proc["Intereses: FSS"])

    output = proc.loc[:, fiscal_metadata[aggregation][fss]]
    columns.set_metadata(output, area="Cuentas fiscales y deuda",
                         currency="UYU", inf_adj="No", index="No",
                         seas_adj="NSA", ts_type="Flujo", cumperiods=1)

    if unit == "gdp":
        output = freqs.rolling(output, periods=12, operation="sum")
        output = convert.pcgdp(output, hifreq=True)
    elif unit == "usd":
        output = convert.usd(output)
    elif unit == "real usd":
        output = convert.real(output, start_date=start_date, end_date=end_date)
        xr = nxr.get(update="nxr.csv", revise_rows=6, save="nxr.csv")
        output = output.divide(xr[start_date:end_date].mean()[3])
        columns.set_metadata(output, currency="USD")
    elif unit == "real":
        output = convert.real(output, start_date=start_date, end_date=end_date)

    if seas_adj in ["trend", "seas"] and unit != "gdp" and cum == 1:
        output_trend, output_seasadj = seasonal.decompose(output, trading=True,
                                                          outlier=True)
        if seas_adj == "trend":
            output = output_trend
        elif seas_adj == "seas":
            output = output_seasadj
        else:
            print("Only 'trend', 'seas' and None are "
                  "possible values for 'seas_adj'")

    if cum != 1:
        output = freqs.rolling(output, periods=cum, operation="sum")

    return output


def labor_mkt(seas_adj: Union[str, None] = "trend"):
    """Get labor market data.

    Allow choosing seasonal adjustment.

    Parameters
    ----------
    seas_adj :
        Allowed strings: 'trend' or 'seas'.

    Returns
    -------
    dataframe : Pandas dataframe

    """
    data = labor.get(update="labor.csv", revise_rows=6,
                     save="labor.csv", force_update=False)
    output = data

    if seas_adj in ["trend", "seas"]:
        trend, seasadj = seasonal.decompose(data, trading=True, outlier=True)
        if seas_adj == "trend":
            output = pd.concat([data, trend], axis=1)
        elif seas_adj == "seas":
            output = pd.concat([data, seasadj], axis=1)

    return output


def nat_accounts(supply: bool = True, real: bool = True, index: bool = False,
                 seas_adj: bool = True, usd: bool = False, cum: int = 1,
                 cust_seas_adj: Union[str, None] = None,
                 variation: Union[str, None] = None):
    """Get national accounts data.

    Attempt to find one of the available data tables with the selected
    combination of parameters (`supply`, `real`. `index` and `seas_adj`).

    Parameters
    ----------
    supply : bool (default is True)
        Supply or demand side.
    real : bool (default is True)
        Constant or current.
    index : bool (default is False)
        Base 100 index or not.
    seas_adj : bool (default is True)
        Seasonally adjusted or not.
    usd : bool (default is False)
        If True, convert to USD.
    cum : int (default is 1)
        How many periods to accumulate for rolling sums.
    cust_seas_adj : str or None (default is None)
        Allowed strings: 'trend' or 'seas'.
    variation : str or None (default is None)
        Type of percentage change to calculate. Can be 'last', 'inter' or
        'annual'.

    Returns
    -------
    dataframe : Pandas dataframe

    Raises
    ------
    KeyError
        If the combined parameters do not correspond to an available table.

    """
    data = national_accounts.get(update=True, revise_rows=4,
                                 save=True, force_update=False)

    search_terms = []
    if supply is True:
        search_terms += ["ind"]
    else:
        search_terms += ["gas"]
    if real is True:
        search_terms += ["con"]
    else:
        search_terms += ["cur"]
    if index is True:
        search_terms += ["idx"]
    if seas_adj is True:
        search_terms += ["sa"]
    else:
        search_terms += ["nsa"]

    table = f"na_{'_'.join(search_terms)}"

    try:
        output = data[table]
    except KeyError:
        print("No available tables with selected parameters.")
        return

    if usd is True:
        output = convert.usd(output)

    if cust_seas_adj is not None and seas_adj is False and cum == 1:
        trend, seasadj = seasonal.decompose(output, trading=True, outlier=True)
        if cust_seas_adj == "trend":
            output = trend
        elif cust_seas_adj == "seas":
            output = seasadj
        else:
            print("Only 'trend', 'seas' and None are "
                  "possible values for 'seas_adj'")

    if cum != 1:
        output = freqs.rolling(output, periods=cum, operation="sum")

    if variation in ["last", "inter", "annual"]:
        output = variations.chg_diff(output, operation="chg",
                                     period_op=variation)
    elif variation is not None:
        print("Only 'last', 'inter' and 'annual' are "
              "possible values for 'variation'")

    return output



import pandas as pd

from econuy.processing import freqs, columns


def chg_diff(df: pd.DataFrame, operation: str = "chg",
             period_op: str = "last"):
    """Wrapper for the pct_change and diff Pandas methods.

    Calculate percentage change or difference for dataframes. The `period`
    argument takes into account the frequency of the dataframe, i.e.,
    `inter` (for interannual) will calculate pct change/differences with
    `periods=4` for quarterly frequency, but `periods=12` for monthly
    frequency.

    Parameters
    ----------
    df : Pandas dataframe
    operation : str (default is 'chg')
        'chg' for percent change or 'diff' for differences.
    period_op : str (default is 'last')
        Period with which to calculate change or difference. 'last' for
        previous period (last month for monthly data), 'inter' for same period
        last year, 'annual' for same period last year but taking annual
        averages/sums.

    Returns
    -------
    output : Pandas dataframe

    Raises
    ------
    ValueError
        If the dataframe is not of frequency M (month end), Q (quarter end) or
        A (year end).

    """
    inferred_freq = pd.infer_freq(df.index)

    type_change = {"last":
                   {"chg": [lambda x: x.pct_change(periods=1),
                            "% variación"],
                    "diff": [lambda x: x.diff(periods=1, min_periods=1),
                             "Cambio"]},
                   "inter":
                   {"chg": [lambda x: x.pct_change(periods=last_year),
                            "% variación interanual"],
                    "diff": [lambda x: x.diff(periods=last_year),
                             "Cambio interanual"]},
                   "annual":
                   {"chg": [lambda x: x.pct_change(periods=last_year),
                            "% variación anual"],
                    "diff": [lambda x: x.diff(periods=last_year),
                             "Cambio anual"]}}

    if inferred_freq == "M":
        last_year = 12
    elif inferred_freq == "Q" or inferred_freq == "Q-DEC":
        last_year = 4
    elif inferred_freq == "A":
        last_year = 1
    else:
        raise ValueError("The dataframe needs to have a frequency of M "
                         "(month end), Q (quarter end) or A (year end)")

    if period_op == "annual":

        if df.columns.get_level_values("Tipo")[0] == "Stock":
            output = df.apply(type_change[period_op][operation][0]).multiply(100)
        else:
            output = freqs.rolling(df, operation="sum")
            output = output.apply(type_change[period_op][operation][0]).multiply(100)

        columns.set_metadata(output, index=type_change[period_op][operation][1])

    else:
        output = df.apply(type_change[period_op][operation][0]).multiply(100)
        columns.set_metadata(output, index=type_change[period_op][operation][1])

    return output



import os
import platform

import pandas as pd
import numpy as np
from statsmodels.api import tsa
from statsmodels.tools.sm_exceptions import X13Error

from econuy.config import ROOT_DIR
from econuy.processing import columns

X13_PATH = os.path.join(ROOT_DIR, "resources", "x13as")

if platform.system() == "Windows":
    X13_PATH = X13_PATH + ".exe"


def decompose(df: pd.DataFrame, trading: bool = True, outlier: bool = True):
    """Apply X13 decomposition. Return trend and seasonally adjusted dataframe.

    Decompose the series in a Pandas dataframe using the US Census X13
    methodology. Will try different combinations of the `trading` and `outlier`
    arguments if an X13 error is raised. Requires having the X13 binary in the
    `resources` folder. Please refer to the README for instructions on where
    to get this binary.

    Parameters
    ----------
    df : Pandas dataframe
    trading : bool (default is True)
        Whether to automatically detect trading days.
    outlier : bool (default is True)
        Whether to automatically detect outliers.

    Returns
    -------
    trend, seas_adj : Pandas dataframe
        Dataframes of the same shape of the input dataframe, containing the
        trend component and the seasonally adjusted series.

    """
    if os.path.isfile(X13_PATH) is False:
        print("X13 binary missing. Place the relevant binary for your system"
              "within the 'resources' directory. Please refer to the README"
              "for instructions on where to get binaries for Windows and Unix,"
              "and how to compile it for macOS.")
        return

    df_proc = df.copy()
    old_columns = df_proc.columns
    df_proc.columns = df_proc.columns.get_level_values(level=0)
    df_proc.index = pd.to_datetime(df_proc.index, errors="coerce")

    trends = []
    seas_adjs = []
    for column in range(len(df_proc.columns)):
        series = df_proc.iloc[:, column].dropna()
        try:
            decomposition = tsa.x13_arima_analysis(
                series, outlier=outlier, trading=trading, forecast_years=0,
                x12path=X13_PATH, prefer_x13=True
            )
            trend = decomposition.trend.reindex(df_proc.index)
            seas_adj = decomposition.seasadj.reindex(df_proc.index)

        except X13Error:
            if outlier is True:
                try:
                    print(f"X13 error found while processing "
                          f"'{df_proc.columns[column]}' with selected "
                          f"parameters. Trying with outlier=False...")
                    return decompose(df=df, outlier=False)

                except X13Error:
                    try:
                        print(f"X13 error found while processing "
                              f"'{df_proc.columns[column]}' with "
                              f"trading=True. Trying with trading=False...")
                        return decompose(df=df, outlier=False,
                                         trading=False)

                    except X13Error:
                        print(f"X13 error found while processing "
                              f"'{df_proc.columns[column]}'. "
                              f"Filling with nan.")
                        trend = pd.Series(np.nan, index=df_proc.index)
                        seas_adj = pd.Series(np.nan, index=df_proc.index)

            elif trading is True:
                try:
                    print(f"X13 error found while processing "
                          f"'{df_proc.columns[column]}' "
                          f"with trading=True. Trying with "
                          f"trading=False...")
                    return decompose(df=df, trading=False)

                except X13Error:
                    print(f"X13 error found while processing "
                          f"'{df_proc.columns[column]}'. Filling with nan.")
                    trend = pd.Series(np.nan, index=df_proc.index)
                    seas_adj = pd.Series(np.nan, index=df_proc.index)

            else:
                try:
                    return decompose(df=df)

                except X13Error:
                    print(f"X13 error found while processing "
                          f"'{df_proc.columns[column]}'. "
                          f"Filling with nan.")
                    trend = pd.Series(np.nan, index=df_proc.index)
                    seas_adj = pd.Series(np.nan, index=df_proc.index)

        trends.append(trend)
        seas_adjs.append(seas_adj)

    trends = pd.concat(trends, axis=1)
    seas_adjs = pd.concat(seas_adjs, axis=1)

    trends.columns = old_columns
    seas_adjs.columns = old_columns

    columns.set_metadata(trends, seas_adj="Tendencia")

    columns.set_metadata(seas_adjs, seas_adj="SA")

    return trends, seas_adjs
